package org.ids_mannheim;

import picocli.CommandLine;

import java.io.File;
import java.io.PrintStream;
import java.nio.file.AccessDeniedException;
import java.nio.file.FileAlreadyExistsException;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Locale;
import java.util.concurrent.*;
import java.util.logging.FileHandler;
import java.util.logging.Logger;
import java.util.logging.SimpleFormatter;
import java.util.stream.Collectors;
import java.util.stream.IntStream;


@CommandLine.Command(mixinStandardHelpOptions = true,
        name = "totalngram", description = "add ngram counts from KorAP-XML or CoNLL-U files")

public class TotalNGram implements Callable<Integer> {
    private static final int MAX_THREADS = Runtime.getRuntime().availableProcessors() * 2 / 3;
    static private final Logger logger;

    static {
        String path = TotalNGram.class.getClassLoader()
                .getResource("logging.properties")
                .getFile();
        System.setProperty("java.util.logging.config.file", path);
        logger = Logger.getLogger(TotalNGram.class.getSimpleName());
    }

    @CommandLine.Parameters(arity = "1..*", description = "input files")
    private final ArrayList<String> inputFiles = new ArrayList<>();
    @CommandLine.Option(names = {"-L",
            "--log-file"}, defaultValue = "totalngram.log", description = "log file name (default: ${DEFAULT-VALUE})")
    String logFileName;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"-o",
            "--output-file"}, description = "Output file (default: ${DEFAULT-VALUE})")
    String output_fillename = "-";

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"--force"}, description = "Force overwrrite (default: ${DEFAULT-VALUE})")
    boolean force_overwrite = false;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"-P",
            "--max-procs"}, description = "Run up to max-procs processes at a time (default: ${DEFAULT-VALUE})")
    int max_threads = MAX_THREADS;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"-p",
            "--worker-pool"}, description = "Run preprocessing on extern hosts, e.g. '10*local,5*host1,3*smith@host2' (default: ${DEFAULT-VALUE})")
    String worker_pool_specification = "";

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"-n",
            "--ngram-size"}, description = "n-gram size (default: ${DEFAULT-VALUE})")
    int ngram_size = 1;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"-f",
            "--folds"}, description = "number of folds (default: ${DEFAULT-VALUE})")
    int FOLDS = 10;
    private Progressbar etaPrinter;

    public TotalNGram() {
    }

    public static void main(String[] args) {
        System.exit(new CommandLine(new TotalNGram()).execute(args));
    }

    @Override
    public Integer call() throws Exception {
        try {
            FileHandler fileHandler = new FileHandler(logFileName);
            fileHandler.setFormatter(new SimpleFormatter());
            //java.util.logging.SimpleFormatter.format="%1$tb %1$td, %1$tY %1$tl:%1$tM:%1$tS %1$Tp %2$s%n%4$s: %5$s%n";
            java.util.Locale.setDefault(Locale.Category.FORMAT, Locale.ROOT);
            logger.addHandler(fileHandler);

        } catch (Exception e) {
        }

        PrintStream output_stream;
        if ((output_fillename == null) || output_fillename.equals("-")) {
            output_stream = System.out;
        } else {
            File f = new File(output_fillename);
            try {
                Files.createFile(f.toPath());
            } catch (AccessDeniedException e) {
                System.err.println("ERROR: Cannot write output file '" + output_fillename + "'");
                System.exit(-1);
            } catch (FileAlreadyExistsException e) {
                if (!force_overwrite) {
                    System.err.println("ERROR: '" + output_fillename + "' already exits. Use --force to overwrite.");
                    System.exit(-1);
                }
            }
            output_stream = new PrintStream(f);
        }

        FoldedEntry.setFolds(FOLDS);
        ConcurrentHashMap<String, FoldedEntry> map = new ConcurrentHashMap<>();
        long totalFilesSizes = inputFiles.parallelStream().mapToLong(fname -> new File(fname).length()).sum();
        etaPrinter = new Progressbar(totalFilesSizes);
        BlockingQueue<Integer> queue = new LinkedBlockingQueue<>(inputFiles.size());
        ExecutorService es = Executors.newCachedThreadPool();
        WorkerNodePool workerNodePool = new WorkerNodePool(worker_pool_specification);
        if (!worker_pool_specification.equals("")) {
            max_threads = workerNodePool.size;
        }
        int threads = Math.min(max_threads, inputFiles.size());
        IntStream.range(0, threads).forEach(unused -> es.execute(new Worker(queue, inputFiles, ngram_size, FOLDS, map, workerNodePool, etaPrinter, logger)));
        queue.addAll(IntStream.range(0, inputFiles.size()).boxed().collect(Collectors.toList()));
        IntStream.range(0, threads).forEach(unused -> {
            try {
                queue.put(-1);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        });
        es.shutdown();
        boolean finished = es.awaitTermination(120, TimeUnit.HOURS);
        etaPrinter.finish();
        logger.info("Sorting and writing frequency table.");
        System.err.println("Sorting and writing frequency table.");
        map.entrySet().parallelStream()
                .sorted(new FreqListEntryComparator<>())
                .forEachOrdered(entry -> output_stream.println(entry.getKey() + entry.getValue().toString()));
        logger.info("Calculating column sums.");
        System.err.println("Calculating column sums.");
        IntStream.rangeClosed(1, FOLDS)
                .parallel()
                .forEachOrdered(i -> output_stream.print("\t" + map.values()
                        .parallelStream().mapToLong(e -> Integer.toUnsignedLong(e.count.get(i))).sum()));
        output_stream.println("\t" + map.values().parallelStream().mapToLong(e -> Integer.toUnsignedLong(e.count.get(0))).sum());
        return null;
    }
}
